@@ -10,6 +10,7 @@
 #define INCLUDED_LOOLKIT_HPP
 
 void lokit_main(const std::string& childRoot,
+                const std::string& jailId,
                 const std::string& sysTemplate,
                 const std::string& loTemplate,
                 const std::string& loSubPath,
@@ -19,7 +20,7 @@ void lokit_main(const std::string& childRoot,
 
 bool globalPreinit(const std::string& loTemplate);
 /// Wrapper around private Document::ViewCallback().
-void documentViewCallback(const int nType, const char* pPayload, void* pData);
+void documentViewCallback(const int type, const char* p, void* data);
 
 class IDocumentManager;
 
@@ -38,8 +39,17 @@ struct UserInfo
 {
     std::string userid;
     std::string username;
+    bool isReadOnly;
 };
 
+/// Check the ForkCounter, and if non-zero, fork more of them accordingly.
+/// @param limit If non-zero, set the ForkCounter to this limit.
+void forkLibreOfficeKit(const std::string& childRoot,
+                        const std::string& sysTemplate,
+                        const std::string& loTemplate,
+                        const std::string& loSubPath,
+                        int limit = 0);
+
 #endif
 
 /* vim:set shiftwidth=4 softtabstop=4 expandtab: */
