@@ -13,50 +13,77 @@
 #include "Session.hpp"
 #include "Storage.hpp"
 #include "MessageQueue.hpp"
-
+#include "SenderQueue.hpp"
+#include "DocumentBroker.hpp"
 #include <Poco/URI.h>
 
 class DocumentBroker;
-class PrisonerSession;
 
 /// Represents a session to a LOOL client, in the WSD process.
-class ClientSession final : public LOOLSession, public std::enable_shared_from_this<ClientSession>
+class ClientSession final : public Session, public std::enable_shared_from_this<ClientSession>
 {
 public:
     ClientSession(const std::string& id,
-                  const std::shared_ptr<LOOLWebSocket>& ws,
                   const std::shared_ptr<DocumentBroker>& docBroker,
                   const Poco::URI& uriPublic,
                   const bool isReadOnly = false);
 
     virtual ~ClientSession();
 
-    void setReadOnly();
-    bool isReadOnly() const { return _isReadOnly; }
+    void handleIncomingMessage(SocketDisposition &) override;
 
-    /// Create and connect Prisoner Session between DocumentBroker and us.
-    void bridgePrisonerSession();
-    std::shared_ptr<PrisonerSession> getPeer() const { return _peer; }
+    void setReadOnly() override;
+
+    /// Returns true if this session is added to a DocBroker.
+    bool isAttached() const { return _isAttached; }
+    void setAttached() { _isAttached = true; }
+
+    /// Returns true if this session has loaded a view (i.e. we got status message).
+    bool isViewLoaded() const { return _isViewLoaded; }
+    void setViewLoaded() { _isViewLoaded = true; }
 
     const std::string getUserId() const { return _userId; }
+    const std::string getUserName() const {return _userName; }
     void setUserId(const std::string& userId) { _userId = userId; }
     void setUserName(const std::string& userName) { _userName = userName; }
     void setDocumentOwner(const bool documentOwner) { _isDocumentOwner = documentOwner; }
     bool isDocumentOwner() const { return _isDocumentOwner; }
 
-    /**
-     * Return the URL of the saved-as document when it's ready. If called
-     * before it's ready, the call blocks till then.
-     */
-    std::string getSaveAsUrl(const unsigned timeoutMs)
+    /// Handle kit-to-client message.
+    bool handleKitToClientMessage(const char* data, const int size);
+
+    // sendTextFrame that takes std::string and string literal.
+    using Session::sendTextFrame;
+
+    bool sendBinaryFrame(const char* buffer, int length) override
     {
-        const auto payload = _saveAsQueue.get(timeoutMs);
-        return std::string(payload.data(), payload.size());
+        auto payload = std::make_shared<Message>(buffer, length, Message::Dir::Out);
+        enqueueSendMessage(payload);
+        return true;
     }
 
-    void setSaveAsUrl(const std::string& url)
+    bool sendTextFrame(const char* buffer, const int length) override
     {
-        _saveAsQueue.put(url);
+        auto payload = std::make_shared<Message>(buffer, length, Message::Dir::Out);
+        enqueueSendMessage(payload);
+        return true;
+    }
+
+    void enqueueSendMessage(const std::shared_ptr<Message>& data)
+    {
+        const auto docBroker = _docBroker.lock();
+        // If in the correct thread - no need for wakeups.
+        if (docBroker)
+            docBroker->assertCorrectThread();
+
+        LOG_TRC(getName() << " enqueueing client message " << data->id());
+        _senderQueue.enqueue(data);
+    }
+
+    /// Set the save-as socket which is used to send convert-to results.
+    void setSaveAsSocket(const std::shared_ptr<StreamSocket>& socket)
+    {
+        _saveAsSocket = socket;
     }
 
     std::shared_ptr<DocumentBroker> getDocumentBroker() const { return _docBroker.lock(); }
@@ -76,54 +103,64 @@ public:
     void setWopiFileInfo(std::unique_ptr<WopiStorage::WOPIFileInfo>& wopiFileInfo) { _wopiFileInfo = std::move(wopiFileInfo); }
 
 private:
+
+    /// SocketHandler: disconnection event.
+    void onDisconnect() override;
+    /// Does SocketHandler: have data or timeouts to setup.
+    int getPollEvents(std::chrono::steady_clock::time_point /* now */,
+                      int & /* timeoutMaxMs */) override;
+    /// SocketHandler: write to socket.
+    void performWrites() override;
+
     virtual bool _handleInput(const char* buffer, int length) override;
 
-    bool loadDocument(const char* buffer, int length, Poco::StringTokenizer& tokens,
+    bool loadDocument(const char* buffer, int length, const std::vector<std::string>& tokens,
                       const std::shared_ptr<DocumentBroker>& docBroker);
     bool getStatus(const char* buffer, int length,
                    const std::shared_ptr<DocumentBroker>& docBroker);
-    bool getCommandValues(const char* buffer, int length, Poco::StringTokenizer& tokens,
+    bool getCommandValues(const char* buffer, int length, const std::vector<std::string>& tokens,
                           const std::shared_ptr<DocumentBroker>& docBroker);
-    bool getPartPageRectangles(const char* buffer, int length,
-                               const std::shared_ptr<DocumentBroker>& docBroker);
-
-    bool sendTile(const char* buffer, int length, Poco::StringTokenizer& tokens,
+    bool sendTile(const char* buffer, int length, const std::vector<std::string>& tokens,
                   const std::shared_ptr<DocumentBroker>& docBroker);
-    bool sendCombinedTiles(const char* buffer, int length, Poco::StringTokenizer& tokens,
+    bool sendCombinedTiles(const char* buffer, int length, const std::vector<std::string>& tokens,
                            const std::shared_ptr<DocumentBroker>& docBroker);
 
-    bool sendFontRendering(const char* buffer, int length, Poco::StringTokenizer& tokens,
+    bool sendFontRendering(const char* buffer, int length, const std::vector<std::string>& tokens,
                            const std::shared_ptr<DocumentBroker>& docBroker);
 
     bool forwardToChild(const std::string& message,
                         const std::shared_ptr<DocumentBroker>& docBroker);
 
+    bool forwardToClient(const std::shared_ptr<Message>& payload);
+
     /// Returns true if given message from the client should be allowed or not
     /// Eg. in readonly mode only few messages should be allowed
     bool filterMessage(const std::string& msg) const;
 
+    void dumpState(std::ostream& os) override;
+
 private:
     std::weak_ptr<DocumentBroker> _docBroker;
 
     /// URI with which client made request to us
     const Poco::URI _uriPublic;
 
-    /// Whether the session is opened as readonly
-    bool _isReadOnly;
-
     /// Whether this session is the owner of currently opened document
     bool _isDocumentOwner;
 
-    /// Our peer that connects us to the child.
-    std::shared_ptr<PrisonerSession> _peer;
+    /// The socket to which the converted (saveas) doc is sent.
+    std::shared_ptr<StreamSocket> _saveAsSocket;
 
-    /// Store URLs of completed 'save as' documents.
-    MessageQueue _saveAsQueue;
+    /// If we are added to a DocBroker.
+    bool _isAttached;
+
+    /// If we have loaded a view.
+    bool _isViewLoaded;
 
-    int _loadPart;
-    
     /// Wopi FileInfo object
     std::unique_ptr<WopiStorage::WOPIFileInfo> _wopiFileInfo;
+
+    SenderQueue<std::shared_ptr<Message>> _senderQueue;
 };
 
 #endif
