@@ -1150,6 +1150,8 @@ bool LOOLWSD::createForKit()
 #else
     LOG_INF("Creating new forkit process.");
 
+    std::unique_lock<std::mutex> newChildrenLock(NewChildrenMutex);
+
     Process::Args args;
     args.push_back("--losubpath=" + std::string(LO_JAIL_SUBPATH));
     args.push_back("--systemplate=" + SysTemplate);
@@ -1174,10 +1176,6 @@ bool LOOLWSD::createForKit()
         args.push_back("--nocaps");
     }
 
-    // If we're recovering forkit, don't allow processing new requests.
-    std::unique_lock<std::mutex> docBrokersLock(DocBrokersMutex);
-    std::unique_lock<std::mutex> newChildrenLock(NewChildrenMutex);
-
     // Always reap first, in case we haven't done so yet.
     int status;
     waitpid(ForKitProcId, &status, WUNTRACED | WNOHANG);
@@ -1208,6 +1206,7 @@ bool LOOLWSD::createForKit()
     // Init the Admin manager
     Admin::instance().setForKitPid(ForKitProcId);
 
+    rebalanceChildren(LOOLWSD::NumPreSpawnedChildren - 1);
     return ForKitProcId != -1;
 #endif
 }
@@ -2495,9 +2494,6 @@ int LOOLWSD::innerMain()
     }
 #endif
 
-    // Prefork initial children
-    prespawnChildren();
-
     // Start the server.
     srv.start(ClientPortNumber);
 
