@@ -7,25 +7,36 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  */
 
-#include "Storage.hpp"
 #include "config.h"
 
+#include "Storage.hpp"
+
 #include <algorithm>
 #include <cassert>
 #include <fstream>
 #include <string>
 
+#include <Poco/DateTime.h>
+#include <Poco/DateTimeParser.h>
+#include <Poco/Exception.h>
 #include <Poco/JSON/Object.h>
 #include <Poco/JSON/Parser.h>
 #include <Poco/Net/DNS.h>
-#include <Poco/Exception.h>
 #include <Poco/Net/HTTPClientSession.h>
 #include <Poco/Net/HTTPRequest.h>
 #include <Poco/Net/HTTPResponse.h>
 #include <Poco/Net/HTTPSClientSession.h>
+#include <Poco/Net/NameValueCollection.h>
 #include <Poco/Net/NetworkInterface.h>
 #include <Poco/Net/SSLManager.h>
 #include <Poco/StreamCopier.h>
+#include <Poco/Timestamp.h>
+
+// For residual Poco SSL usage.
+#include <Poco/Net/Context.h>
+#include <Poco/Net/SSLManager.h>
+#include <Poco/Net/AcceptCertificateHandler.h>
+#include <Poco/Net/KeyConsoleHandler.h>
 
 #include "Auth.hpp"
 #include "Common.hpp"
@@ -94,6 +105,24 @@ void StorageBase::initialize()
             }
         }
     }
+
+#if ENABLE_SSL
+    // FIXME: should use our own SSL socket implementation here.
+    Poco::Crypto::initializeCrypto();
+    Poco::Net::initializeSSL();
+
+    // Init client
+    Poco::Net::Context::Params sslClientParams;
+
+    // TODO: Be more strict and setup SSL key/certs for remove server and us
+    sslClientParams.verificationMode = Poco::Net::Context::VERIFY_NONE;
+
+    Poco::SharedPtr<Poco::Net::PrivateKeyPassphraseHandler> consoleClientHandler = new Poco::Net::KeyConsoleHandler(false);
+    Poco::SharedPtr<Poco::Net::InvalidCertificateHandler> invalidClientCertHandler = new Poco::Net::AcceptCertificateHandler(false);
+
+    Poco::Net::Context::Ptr sslClientContext = new Poco::Net::Context(Poco::Net::Context::CLIENT_USE, sslClientParams);
+    Poco::Net::SSLManager::instance().initializeClient(consoleClientHandler, invalidClientCertHandler, sslClientContext);
+#endif
 }
 
 bool isLocalhost(const std::string& targetHost)
@@ -105,14 +134,14 @@ bool isLocalhost(const std::string& targetHost)
     }
     catch (const Poco::Exception& exc)
     {
-        Log::warn("Poco::Net::DNS::resolveOne(\"" + targetHost + "\") failed: " + exc.displayText());
+        LOG_WRN("Poco::Net::DNS::resolveOne(\"" << targetHost << "\") failed: " << exc.displayText());
         try
         {
             targetAddress = Poco::Net::IPAddress(targetHost).toString();
         }
         catch (const Poco::Exception& exc1)
         {
-            Log::warn("Poco::Net::IPAddress(\"" + targetHost + "\") failed: " + exc1.displayText());
+            LOG_WRN("Poco::Net::IPAddress(\"" << targetHost << "\") failed: " << exc1.displayText());
         }
     }
 
@@ -191,15 +220,12 @@ std::unique_ptr<LocalStorage::LocalFileInfo> LocalStorage::getLocalFileInfo()
     const auto path = Poco::Path(_uri.getPath());
     LOG_DBG("Getting info for local uri [" << _uri.toString() << "], path [" << path.toString() << "].");
 
-    if (!_fileInfo.isValid())
-    {
-        const auto& filename = path.getFileName();
-        const auto file = Poco::File(path);
-        const auto lastModified = file.getLastModified();
-        const auto size = file.getSize();
+    const auto& filename = path.getFileName();
+    const auto file = Poco::File(path);
+    const auto lastModified = file.getLastModified();
+    const auto size = file.getSize();
 
-        _fileInfo = FileInfo({filename, "localhost", lastModified, size});
-    }
+    _fileInfo = FileInfo({filename, "localhost", lastModified, size});
 
     // Set automatic userid and username
     return std::unique_ptr<LocalStorage::LocalFileInfo>(new LocalFileInfo({"localhost", std::string("Local Host #") + std::to_string(LastLocalStorageId++)}));
@@ -211,21 +237,21 @@ std::string LocalStorage::loadStorageFileToLocal(const std::string& /*accessToke
     const auto filename = Poco::Path(_uri.getPath()).getFileName();
     _jailedFilePath = Poco::Path(getLocalRootPath(), filename).toString();
     LOG_INF("Public URI [" << _uri.getPath() <<
-            "] jailed to [" + _jailedFilePath + "].");
+            "] jailed to [" << _jailedFilePath << "].");
 
     // Despite the talk about URIs it seems that _uri is actually just a pathname here
     const auto publicFilePath = _uri.getPath();
 
-    if (!FileUtil::checkDiskSpace(publicFilePath))
+    if (!FileUtil::checkDiskSpace(_jailedFilePath))
     {
-        throw StorageSpaceLowException("Low disk space for " + publicFilePath);
+        throw StorageSpaceLowException("Low disk space for " + _jailedFilePath);
     }
 
     LOG_INF("Linking " << publicFilePath << " to " << _jailedFilePath);
     if (!Poco::File(_jailedFilePath).exists() && link(publicFilePath.c_str(), _jailedFilePath.c_str()) == -1)
     {
         // Failed
-        Log::warn("link(\"" + publicFilePath + "\", \"" + _jailedFilePath + "\") failed. Will copy.");
+        LOG_WRN("link(\"" << publicFilePath << "\", \"" << _jailedFilePath << "\") failed. Will copy.");
     }
 
     try
@@ -240,13 +266,20 @@ std::string LocalStorage::loadStorageFileToLocal(const std::string& /*accessToke
     }
     catch (const Poco::Exception& exc)
     {
-        Log::error("copyTo(\"" + publicFilePath + "\", \"" + _jailedFilePath + "\") failed: " + exc.displayText());
+        LOG_ERR("copyTo(\"" << publicFilePath << "\", \"" << _jailedFilePath << "\") failed: " << exc.displayText());
         throw;
     }
 
     _isLoaded = true;
     // Now return the jailed path.
-    return Poco::Path(_jailPath, filename).toString();
+#ifndef KIT_IN_PROCESS
+    if (LOOLWSD::NoCapsForKit)
+        return _jailedFilePath;
+    else
+        return Poco::Path(_jailPath, filename).toString();
+#else
+    return _jailedFilePath;
+#endif
 }
 
 StorageBase::SaveResult LocalStorage::saveLocalFileToStorage(const std::string& /*accessToken*/)
@@ -272,11 +305,15 @@ StorageBase::SaveResult LocalStorage::saveLocalFileToStorage(const std::string&
 
 namespace {
 
-static inline
+inline
 Poco::Net::HTTPClientSession* getHTTPClientSession(const Poco::URI& uri)
 {
-    return (LOOLWSD::isSSLEnabled() || LOOLWSD::isSSLTermination()) ? new Poco::Net::HTTPSClientSession(uri.getHost(), uri.getPort(), Poco::Net::SSLManager::instance().defaultClientContext())
-                       : new Poco::Net::HTTPClientSession(uri.getHost(), uri.getPort());
+    // FIXME: if we're configured for http - we can still use an https:// wopi
+    // host surely; of course - the converse is not true / sensible.
+    return (LOOLWSD::isSSLEnabled() || LOOLWSD::isSSLTermination())
+        ? new Poco::Net::HTTPSClientSession(uri.getHost(), uri.getPort(),
+                                            Poco::Net::SSLManager::instance().defaultClientContext())
+        : new Poco::Net::HTTPClientSession(uri.getHost(), uri.getPort());
 }
 
 int getLevenshteinDist(const std::string& string1, const std::string& string2) {
@@ -368,6 +405,25 @@ void setQueryParameter(Poco::URI& uriObject, const std::string& key, const std::
     uriObject.addQueryParameter(key, value);
 }
 
+
+void addStorageDebugCookie(Poco::Net::HTTPRequest& request)
+{
+    (void) request;
+#if ENABLE_DEBUG
+    if (std::getenv("LOOL_STORAGE_COOKIE"))
+    {
+        Poco::Net::NameValueCollection nvcCookies;
+        std::vector<std::string> cookieTokens = LOOLProtocol::tokenize(std::string(std::getenv("LOOL_STORAGE_COOKIE")), ':');
+        if (cookieTokens.size() == 2)
+        {
+            nvcCookies.add(cookieTokens[0], cookieTokens[1]);
+            request.setCookies(nvcCookies);
+            LOG_TRC("Added storage debug cookie [" << cookieTokens[0] << "=" << cookieTokens[1] << "].");
+        }
+    }
+#endif
+}
+
 } // anonymous namespace
 
 std::unique_ptr<WopiStorage::WOPIFileInfo> WopiStorage::getWOPIFileInfo(const std::string& accessToken)
@@ -376,26 +432,50 @@ std::unique_ptr<WopiStorage::WOPIFileInfo> WopiStorage::getWOPIFileInfo(const st
     Poco::URI uriObject(_uri);
     setQueryParameter(uriObject, "access_token", accessToken);
 
-    Log::debug("Getting info for wopi uri [" + uriObject.toString() + "].");
+    LOG_DBG("Getting info for wopi uri [" << uriObject.toString() << "].");
 
+    std::string resMsg;
     const auto startTime = std::chrono::steady_clock::now();
-    std::unique_ptr<Poco::Net::HTTPClientSession> psession(getHTTPClientSession(uriObject));
-
-    Poco::Net::HTTPRequest request(Poco::Net::HTTPRequest::HTTP_GET, uriObject.getPathAndQuery(), Poco::Net::HTTPMessage::HTTP_1_1);
-    request.set("User-Agent", "LOOLWSD WOPI Agent");
-    psession->sendRequest(request);
-
-    Poco::Net::HTTPResponse response;
-
-    std::istream& rs = psession->receiveResponse(response);
-    auto logger = Log::trace();
-    logger << "WOPI::CheckFileInfo header for URI [" << uriObject.toString() << "]:\n";
-    for (auto& pair : response)
+    std::chrono::duration<double> callDuration(0);
+    try
     {
-        logger << '\t' + pair.first + ": " + pair.second << " / ";
-    }
+        std::unique_ptr<Poco::Net::HTTPClientSession> psession(getHTTPClientSession(uriObject));
 
-    logger << Log::end;
+        Poco::Net::HTTPRequest request(Poco::Net::HTTPRequest::HTTP_GET, uriObject.getPathAndQuery(), Poco::Net::HTTPMessage::HTTP_1_1);
+        request.set("User-Agent", WOPI_AGENT_STRING);
+        addStorageDebugCookie(request);
+        psession->sendRequest(request);
+
+        Poco::Net::HTTPResponse response;
+        std::istream& rs = psession->receiveResponse(response);
+        callDuration = (std::chrono::steady_clock::now() - startTime);
+
+        auto logger = Log::trace();
+        if (logger.enabled())
+        {
+            logger << "WOPI::CheckFileInfo header for URI [" << uriObject.toString() << "]:\n";
+            for (const auto& pair : response)
+            {
+                logger << '\t' << pair.first << ": " << pair.second << " / ";
+            }
+
+            LOG_END(logger);
+        }
+
+        if (response.getStatus() != Poco::Net::HTTPResponse::HTTP_OK)
+        {
+            LOG_ERR("WOPI::CheckFileInfo failed with " << response.getStatus() << ' ' << response.getReason());
+            throw StorageConnectionException("WOPI::CheckFileInfo failed");
+        }
+
+        Poco::StreamCopier::copyToString(rs, resMsg);
+    }
+    catch(const Poco::Exception& pexc)
+    {
+        LOG_ERR("Cannot get file info from WOPI storage uri [" << uriObject.toString() << "]. Error: " << pexc.displayText() <<
+                (pexc.nested() ? " (" + pexc.nested()->displayText() + ")" : ""));
+        throw;
+    }
 
     // Parse the response.
     std::string filename;
@@ -409,12 +489,12 @@ std::unique_ptr<WopiStorage::WOPIFileInfo> WopiStorage::getWOPIFileInfo(const st
     bool hidePrintOption = false;
     bool hideSaveOption = false;
     bool hideExportOption = false;
-    std::string resMsg;
-    Poco::StreamCopier::copyToString(rs, resMsg);
+    bool disablePrint = false;
+    bool disableExport = false;
+    bool disableCopy = false;
+    std::string lastModifiedTime;
 
-    const auto endTime = std::chrono::steady_clock::now();
-    const std::chrono::duration<double> callDuration = (endTime - startTime);
-    Log::debug("WOPI::CheckFileInfo returned: " + resMsg + ". Call duration: " + std::to_string(callDuration.count()) + "s");
+    LOG_DBG("WOPI::CheckFileInfo returned: " << resMsg << ". Call duration: " << callDuration.count() << "s");
     const auto index = resMsg.find_first_of('{');
     if (index != std::string::npos)
     {
@@ -433,17 +513,42 @@ std::unique_ptr<WopiStorage::WOPIFileInfo> WopiStorage::getWOPIFileInfo(const st
         getWOPIValue(object, "HideSaveOption", hideSaveOption);
         getWOPIValue(object, "HideExportOption", hideExportOption);
         getWOPIValue(object, "EnableOwnerTermination", enableOwnerTermination);
+        getWOPIValue(object, "DisablePrint", disablePrint);
+        getWOPIValue(object, "DisableExport", disableExport);
+        getWOPIValue(object, "DisableCopy", disableCopy);
+        getWOPIValue(object, "LastModifiedTime", lastModifiedTime);
     }
     else
-        Log::error("WOPI::CheckFileInfo is missing JSON payload");
-
-    if (!_fileInfo.isValid())
     {
-        // WOPI doesn't support file last modified time.
-        _fileInfo = FileInfo({filename, ownerId, Poco::Timestamp(), size});
+        LOG_ERR("WOPI::CheckFileInfo failed and no JSON payload returned. Access denied.");
+        throw UnauthorizedRequestException("Access denied. WOPI::CheckFileInfo failed on: " + uriObject.toString());
     }
 
-    return std::unique_ptr<WopiStorage::WOPIFileInfo>(new WOPIFileInfo({userId, userName, canWrite, postMessageOrigin, hidePrintOption, hideSaveOption, hideExportOption, enableOwnerTermination, callDuration}));
+    Poco::Timestamp modifiedTime = Poco::Timestamp::fromEpochTime(0);
+    if (lastModifiedTime != "")
+    {
+        Poco::DateTime dateTime;
+        int timeZoneDifferential;
+        bool valid = false;
+        try
+        {
+            Poco::DateTimeParser::parse(Poco::DateTimeFormat::ISO8601_FRAC_FORMAT, lastModifiedTime, dateTime, timeZoneDifferential);
+            valid = true;
+        }
+        catch (const Poco::SyntaxException& exc)
+        {
+            LOG_WRN("LastModifiedTime property [" << lastModifiedTime << "] was invalid format: " << exc.displayText() <<
+                    (exc.nested() ? " (" + exc.nested()->displayText() + ")" : ""));
+        }
+        if (valid)
+        {
+            modifiedTime = dateTime.timestamp();
+        }
+    }
+
+    _fileInfo = FileInfo({filename, ownerId, modifiedTime, size});
+
+    return std::unique_ptr<WopiStorage::WOPIFileInfo>(new WOPIFileInfo({userId, userName, canWrite, postMessageOrigin, hidePrintOption, hideSaveOption, hideExportOption, enableOwnerTermination, disablePrint, disableExport, disableCopy, callDuration}));
 }
 
 /// uri format: http://server/<...>/wopi*/files/<id>/content
@@ -454,43 +559,63 @@ std::string WopiStorage::loadStorageFileToLocal(const std::string& accessToken)
     Poco::URI uriObject(_uri);
     uriObject.setPath(uriObject.getPath() + "/contents");
     setQueryParameter(uriObject, "access_token", accessToken);
-    Log::debug("Wopi requesting: " + uriObject.toString());
+    LOG_DBG("Wopi requesting: " << uriObject.toString());
 
     const auto startTime = std::chrono::steady_clock::now();
-    std::unique_ptr<Poco::Net::HTTPClientSession> psession(getHTTPClientSession(uriObject));
-
-    Poco::Net::HTTPRequest request(Poco::Net::HTTPRequest::HTTP_GET, uriObject.getPathAndQuery(), Poco::Net::HTTPMessage::HTTP_1_1);
-    request.set("User-Agent", "LOOLWSD WOPI Agent");
-    psession->sendRequest(request);
-
-    Poco::Net::HTTPResponse response;
-    std::istream& rs = psession->receiveResponse(response);
-
-    auto logger = Log::trace();
-    logger << "WOPI::GetFile header for URI [" << uriObject.toString() << "]:\n";
-    for (auto& pair : response)
+    try
     {
-        logger << '\t' + pair.first + ": " + pair.second << " / ";
+        std::unique_ptr<Poco::Net::HTTPClientSession> psession(getHTTPClientSession(uriObject));
+
+        Poco::Net::HTTPRequest request(Poco::Net::HTTPRequest::HTTP_GET, uriObject.getPathAndQuery(), Poco::Net::HTTPMessage::HTTP_1_1);
+        request.set("User-Agent", WOPI_AGENT_STRING);
+        addStorageDebugCookie(request);
+        psession->sendRequest(request);
+
+        Poco::Net::HTTPResponse response;
+        std::istream& rs = psession->receiveResponse(response);
+        const std::chrono::duration<double> diff = (std::chrono::steady_clock::now() - startTime);
+        _wopiLoadDuration += diff;
+
+        auto logger = Log::trace();
+        if (logger.enabled())
+        {
+            logger << "WOPI::GetFile header for URI [" << uriObject.toString() << "]:\n";
+            for (const auto& pair : response)
+            {
+                logger << '\t' << pair.first << ": " << pair.second << " / ";
+            }
+
+            LOG_END(logger);
+        }
+
+        if (response.getStatus() != Poco::Net::HTTPResponse::HTTP_OK)
+        {
+            LOG_ERR("WOPI::GetFile failed with " << response.getStatus() << ' ' << response.getReason());
+            throw StorageConnectionException("WOPI::GetFile failed");
+        }
+        else // Successful
+        {
+            _jailedFilePath = Poco::Path(getLocalRootPath(), _fileInfo._filename).toString();
+            std::ofstream ofs(_jailedFilePath);
+            std::copy(std::istreambuf_iterator<char>(rs),
+                      std::istreambuf_iterator<char>(),
+                      std::ostreambuf_iterator<char>(ofs));
+            LOG_INF("WOPI::GetFile downloaded " << getFileSize(_jailedFilePath) << " bytes from [" << uriObject.toString() <<
+                    "] -> " << _jailedFilePath << " in " << diff.count() << "s");
+
+            _isLoaded = true;
+            // Now return the jailed path.
+            return Poco::Path(_jailPath, _fileInfo._filename).toString();
+        }
+    }
+    catch(const Poco::Exception& pexc)
+    {
+        LOG_ERR("Cannot load document from WOPI storage uri [" + uriObject.toString() + "]. Error: " << pexc.displayText() <<
+                (pexc.nested() ? " (" + pexc.nested()->displayText() + ")" : ""));
+        throw;
     }
 
-    logger << Log::end;
-
-    _jailedFilePath = Poco::Path(getLocalRootPath(), _fileInfo._filename).toString();
-    std::ofstream ofs(_jailedFilePath);
-    std::copy(std::istreambuf_iterator<char>(rs),
-              std::istreambuf_iterator<char>(),
-              std::ostreambuf_iterator<char>(ofs));
-    const auto endTime = std::chrono::steady_clock::now();
-    const std::chrono::duration<double> diff = (endTime - startTime);
-    _wopiLoadDuration += diff;
-    const auto size = getFileSize(_jailedFilePath);
-    LOG_INF("WOPI::GetFile downloaded " << size << " bytes from [" << uriObject.toString() <<
-            "] -> " << _jailedFilePath << " in " << diff.count() << "s : " <<
-            response.getStatus() << " " << response.getReason());
-
-    _isLoaded = true;
-    // Now return the jailed path.
-    return Poco::Path(_jailPath, _fileInfo._filename).toString();
+    return "";
 }
 
 StorageBase::SaveResult WopiStorage::saveLocalFileToStorage(const std::string& accessToken)
@@ -501,37 +626,50 @@ StorageBase::SaveResult WopiStorage::saveLocalFileToStorage(const std::string& a
     Poco::URI uriObject(_uri);
     uriObject.setPath(uriObject.getPath() + "/contents");
     setQueryParameter(uriObject, "access_token", accessToken);
-    Log::debug("Wopi posting: " + uriObject.toString());
 
-    std::unique_ptr<Poco::Net::HTTPClientSession> psession(getHTTPClientSession(uriObject));
+    LOG_INF("Uploading URI via WOPI [" << uriObject.toString() << "] from [" << _jailedFilePath + "].");
 
-    Poco::Net::HTTPRequest request(Poco::Net::HTTPRequest::HTTP_POST, uriObject.getPathAndQuery(), Poco::Net::HTTPMessage::HTTP_1_1);
-    request.set("X-WOPI-Override", "PUT");
-    request.setContentType("application/octet-stream");
-    request.setContentLength(size);
-
-    std::ostream& os = psession->sendRequest(request);
-    std::ifstream ifs(_jailedFilePath);
-    Poco::StreamCopier::copyStream(ifs, os);
-
-    Poco::Net::HTTPResponse response;
-    std::istream& rs = psession->receiveResponse(response);
     std::ostringstream oss;
-    Poco::StreamCopier::copyStream(rs, oss);
-
-    LOG_INF("WOPI::PutFile response: " << oss.str());
-    LOG_INF("WOPI::PutFile uploaded " << size << " bytes from [" << _jailedFilePath <<
-            "] -> [" << uriObject.toString() << "]: " <<
-            response.getStatus() << " " << response.getReason());
-
     StorageBase::SaveResult saveResult = StorageBase::SaveResult::FAILED;
-    if (response.getStatus() == Poco::Net::HTTPResponse::HTTP_OK)
+    try
     {
-        saveResult = StorageBase::SaveResult::OK;
+        std::unique_ptr<Poco::Net::HTTPClientSession> psession(getHTTPClientSession(uriObject));
+
+        Poco::Net::HTTPRequest request(Poco::Net::HTTPRequest::HTTP_POST, uriObject.getPathAndQuery(), Poco::Net::HTTPMessage::HTTP_1_1);
+        request.set("X-WOPI-Override", "PUT");
+        request.setContentType("application/octet-stream");
+        request.setContentLength(size);
+        addStorageDebugCookie(request);
+        std::ostream& os = psession->sendRequest(request);
+        std::ifstream ifs(_jailedFilePath);
+        Poco::StreamCopier::copyStream(ifs, os);
+
+        Poco::Net::HTTPResponse response;
+        std::istream& rs = psession->receiveResponse(response);
+        Poco::StreamCopier::copyStream(rs, oss);
+        LOG_INF("WOPI::PutFile response: " << oss.str());
+        LOG_INF("WOPI::PutFile uploaded " << size << " bytes from [" << _jailedFilePath <<
+                "] -> [" << uriObject.toString() << "]: " <<
+                response.getStatus() << " " << response.getReason());
+
+        if (response.getStatus() == Poco::Net::HTTPResponse::HTTP_OK)
+        {
+            saveResult = StorageBase::SaveResult::OK;
+        }
+        else if (response.getStatus() == Poco::Net::HTTPResponse::HTTP_REQUESTENTITYTOOLARGE)
+        {
+            saveResult = StorageBase::SaveResult::DISKFULL;
+        }
+        else if (response.getStatus() == Poco::Net::HTTPResponse::HTTP_UNAUTHORIZED)
+        {
+            saveResult = StorageBase::SaveResult::UNAUTHORIZED;
+        }
     }
-    else if (response.getStatus() == Poco::Net::HTTPResponse::HTTP_REQUESTENTITYTOOLARGE)
+    catch(const Poco::Exception& pexc)
     {
-        saveResult = StorageBase::SaveResult::DISKFULL;
+        LOG_ERR("Cannot save file to WOPI storage uri [" + uriObject.toString() + "]. Error: " << pexc.displayText() <<
+                (pexc.nested() ? " (" + pexc.nested()->displayText() + ")" : ""));
+        saveResult = StorageBase::SaveResult::FAILED;
     }
 
     return saveResult;
